/*

XFILTER - device driver which captures IDL output

xfiltdlm.c

Copyright (C) 1999 Craig B. Markwardt.  All Rights Reserved.

This module includes all of the non-portable code which depends on
individual architectures.  The primary difficulty is getting access to
the addresses of two symbols:

 - IDL_RBtreeSearch_IDByStr - finds and IDL command
 - spro_rbtree

There are three methods I have developed, which range from the fairly
elegant to the downright hideous.  When porting to a new architecture,
one is recommended to try each in order.

You must include your architecture in the appropriate #ifdef in order
to enable it for that method, otherwise XFILTER will fail.

Method 1 - DIRECT SYMBOLS 
The symbols may be available directly.  This is by far the easiest, so
test this first.  If you get "unresolved" symbol errors, then you need
to try methods 2 or 3.

Method 2 - DYNAMIC SYMBOLS

The symbols might be found using the dlopen and dlsym functions.  This
is fairly easy to do, just be sure you put the correct architecture
name to locate the correct share library file.

Method 3 - USE "NM" TO EXTRACT SYMBOL NAMES

For Solaris and IRIX there appears to be no way to link the required
symbols.  Thus I use the "nm" command to pull out the symbol names of
interest.  True, it's slow, but it's better than hard-coding the values.

************************ FINALLY ***************************

You need to supply an algorithm which can return an IDL structure's
tag names.  Under IDL 5.2 and greater, it is possible to do this with
a documented routine.  Otherwise it has to be hard-coded in.  When
porting to a enw architecture, you should try the ones that already
work for known architectures, and hope!

See xfilter_tag_name below.

************************************************************

This file is part of XFILTER.

XFILTER is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

XFILTER is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with XFILTER; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "export.h"

long int **proc_table;
IDL_SYSFUN_DEF *(*proc_search)(unsigned long int *, char *);
char *arch = 0, *rel = 0;
float drel = 0;

/* Method 1 - DIRECT SYMBOLS *************************************** */
/* Works for Linux and SunOS */
#if defined(linux) || defined(sunos)
#define spro_rbtree _IDL_spro_rbtree
extern unsigned long int *spro_rbtree;
extern void * IDL_RBtreeSearch_IDByStr(unsigned long *, char *);
void xfilter_init_symbols1()
{
  proc_table = (long int **) &spro_rbtree;
  proc_search = (IDL_SYSFUN_DEF * (*)(unsigned long int *, char *))
    &IDL_RBtreeSearch_IDByStr;
}
#define xfilt_dlm
#endif

/* Method 2 - DYNAMIC SYMBOLS *************************************** */
/* Works for Alpha */
#if defined(alpha)

/* set your architecture name here, as it appears in $IDL_DIR/bin/bin.XXX  */
#if defined(alpha)
#define hosttype "alpha"
#endif

#include <dlfcn.h>
void xfilter_init_symbols1()
{
  void *lib;
  char pathname[256] = "";

  strcat(pathname, IDL_STRING_STR(&IDL_SysvDir));
  strcat(pathname, "/bin/bin." hosttype "/libidl.so");
  
  lib = dlopen(pathname, RTLD_LAZY);
  if (lib == 0) {
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
	"ERROR: (" hosttype ") shared library libidl.so was not found.");
    exit(1);
  }
  proc_table = (long int **) dlsym(lib, "spro_rbtree");
  if (proc_table == 0) {
    dlclose(lib);
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
	"ERROR: (" hosttype ") library symbol spro_rbtree was not found.");
    exit(1);
  }
  proc_search = dlsym(lib, "IDL_RBtreeSearch_IDByStr");
  if (proc_search == 0) {
    dlclose(lib);
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
	"ERROR: (" hosttype ") library symbol spro_rbtree was not found.");
    exit(1);
  }
  dlclose(lib);

  return;
}
#define xfilt_dlm
#endif

/* Method 3 - USE NM TO EXTRACT SYMBOL NAMES ************************ */
#if defined(sgi) || defined(solaris)
void xfilter_init_symbols1()
{
  long int proc_search_offset = 0;
  long int proc_table_offset  = 0;
  unsigned long int spro_rbtree_offset = 0;
  unsigned long int IDL_RBtreeSearch_IDByStr_offset = 0;
  unsigned long int IDL_ColorMap_offset = 0;
  FILE *p = 0;
  char nm[300] = "";
  char line[300];
  char *debug = getenv("XFILTER_DEBUG");

  /* Reference point is IDL_ColorMap */
  /* proc_search is IDL_RBtreeSearch_IDByStr */
  /* proc_table  is spro_rbtree */

  /* SGI Irix special cases */
#if defined(sgi)
#define hosttype "sgi"
#define nm_command "/usr/bin/nm -B "
#endif

  /* Solaris specific cases, both x86 and sparc */
#if defined(solaris)
#if defined(sparc)
#define hosttype "solaris2.sparc"
#else
#define hosttype "solaris2.x86"
#endif
#define nm_command "/usr/ccs/bin/nm -p -v -x "
#endif

  /* AIX specific cases */
#if defined(aix)
#define hosttype "aix"
#define nm_command "/usr/bin/nm -p -v -x "
#endif

  /* Construct the nm command */
  strcat(nm, nm_command);
  strcat(nm, IDL_STRING_STR(&IDL_SysvDir));
  strcat(nm, "/bin/bin." hosttype "/libidl.so ");
  if (debug) {
    fprintf(stderr, "nm command : %s\n", nm);
  }

  /* Run the command and extract the output */
  p = popen(nm, "r");
  if (p == 0) {
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
		"ERROR: (" hosttype ") nm pipe could not complete.");
    exit(1);
  }
  while (!feof(p)) {
    unsigned long int offset;
    char symname[100];

    if (fgets(line, 299, p) == 0) break;
    line[299] = 0;
    if(sscanf(line, "%lx %*c %s", &offset, symname) != 2)
      continue;

    /* Look for the symbols of interest */
    if (strcmp(symname, "spro_rbtree") == 0)
      spro_rbtree_offset = offset;
    else if (strcmp(symname, "IDL_ColorMap") == 0)
      IDL_ColorMap_offset = offset;
    else if (strcmp(symname, "IDL_RBtreeSearch_IDByStr") == 0)
      IDL_RBtreeSearch_IDByStr_offset = offset;
    else
      continue;

    if (debug) {
      fprintf(stderr, "nm = %s\n", line);
    }
  }
  pclose(p);

  if (debug) {
    printf("sp = %lx\n", spro_rbtree_offset);
    printf("CM = %lx\n", IDL_ColorMap_offset);
    printf("RB = %lx\n", IDL_RBtreeSearch_IDByStr_offset);
  }

  if (spro_rbtree_offset == 0
      || IDL_ColorMap_offset == 0 
      || IDL_RBtreeSearch_IDByStr_offset == 0) {
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
		"ERROR: (" hosttype ") shared library symbols were not found.");
    exit(1);
  }

  proc_table = (long int **) 
    ((char *) &IDL_ColorMap + spro_rbtree_offset - IDL_ColorMap_offset);
  proc_search = (IDL_SYSFUN_DEF * (*)(unsigned long int *, char *))
    ((char *) &IDL_ColorMap + IDL_RBtreeSearch_IDByStr_offset - IDL_ColorMap_offset);

  return;
}
#define xfilt_dlm
#endif

/* Fall-back case */
#ifndef xfilt_dlm
void xfilter_init_symbols1()
{
  IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
	      "ERROR: XFILTER is not compatible with this architecture.");
  exit(1);
}
#endif

/* Main entry to this module ************************************** */
void xfilter_init_symbols()
{
  if (rel == 0) {
    arch = IDL_STRING_STR(&IDL_SysvVersion.arch);
    rel  = IDL_STRING_STR(&IDL_SysvVersion.release);
    drel = atof(rel);
  }
  if (drel < 5.09) {
    IDL_Message(IDL_M_GENERIC, IDL_MSG_LONGJMP, 
		"ERROR: XFILTER is not compatible with this IDL version.");
    exit(1);
  }
  xfilter_init_symbols1();
}

/* Extracting a structure's tag names */
char * xfilter_tag_name(IDL_StructDefPtr sdef, int tag_num)
{
  static int hasfun = -1, symfun = -1;
  if (hasfun < 0) hasfun = (atof(rel) > 5.15); /* Versions 5.2 and greater */
  if (symfun < 0) symfun = (atof(rel) > 5.05); /* Versions 5.1 and 5.1.1   */
    
  /* hasfun indicates that the IDL_StructTagNameByIndex is present */
  if (hasfun) 
    return IDL_StructTagNameByIndex(sdef, tag_num, 0, 0);
  /* symfun indicates that we will try to simulate it. */
  if (!symfun)
    return 0;
  
  if (tag_num < 0) return 0;

  /* Truly gruesome platform independent hacks */
#if defined(__sun__) || defined(sgi)
  return *((char **) (**(char ***) ((char *) sdef + tag_num*40 + 24) + 4));
#endif
#if defined(linux)
  return *((char **) (**(char ***) ((char *) sdef + tag_num*36 + 24) + 4));
#endif
#if defined(alpha)
  return *((char **) (**(char ***) ((char *) sdef + tag_num*56 + 32) + 8));
#endif
  return 0;
}    
    
